(ns dmr.events
  (:require
   [clojure.string :refer [lower-case split join starts-with?]]
   [re-frame.core :as re-frame]
   [dmr.db :as db]
   [day8.re-frame.tracing :refer-macros [fn-traced]]))


(re-frame/reg-event-db
 ::initialize-db
 (fn-traced [_ _]
   db/default-db))

(defn single-space [s]
  (as-> s x
        (split x " ")
        (filter not-empty x)
        (join " " x)))

(defn normalize-cmdline [cmdline]
    (-> cmdline single-space lower-case))

(defn- concatv [m]
    (->> m vals (apply concat)))

(defn autocomplete-suggestions [db current]
  (let [entities (db :entities)
        cur (normalize-cmdline current)
        best-match #(< (count %1) (count %2))]
      (when (not-empty cur)
        (->> (concatv entities)
             (filter #(-> % :name normalize-cmdline (starts-with? cur)))
             (map :name)
             (sort best-match)))))

(re-frame/reg-event-db
 ::update-cmdline
 (fn-traced [db [_ current]]
  (-> db
      (assoc-in [:cmdline :current] current)
      (assoc-in [:cmdline :suggestions] {:options (autocomplete-suggestions db current)
                                         :selected nil}))))

(defn get-using [f coll]
  (->> coll (filter f) first))

(defn execute [db cmdline]
 (let [append-history #(update-in db [:cmdline :history] conj {:input cmdline :output %})
       entities (db :entities)
       lcc (normalize-cmdline cmdline)
       get-using-name (partial get-using #(some-> % :name normalize-cmdline (= lcc)))]
  ; TODO: Replace with the 'match' macro
  (cond
   (empty? cmdline) db
   (= "clear" cmdline) (assoc-in db [:cmdline :history] [])
   (= "Open Game License" cmdline) (append-history (db :license))
   (= "spells" cmdline) (-> entities :spells append-history)
   (= "items" cmdline)  (-> entities :equipment append-history)
   ; TODO: refactor this to avoid 2 lookups (cond -> or)
   (-> entities concatv (get-using-name)) (-> entities concatv get-using-name append-history)
   :else (append-history {:err (str "Unknown command '" cmdline "'")}))))

(re-frame/reg-fx
 :window/scroll-to
 (fn [element]
   (.scrollIntoView element (clj->js {:behavior :smooth}))))

(re-frame/reg-event-fx
 ::scroll-to
 (fn [_ [_ x]]
     {:window/scroll-to x}))

(defn- cycle-history [cs direction]
  (let [history (-> cs :history)
        current (-> cs :current)
        orig (some-> cs :selected-history :orig)
        idx (some-> cs :selected-history :index)
        history-end (- (count history) 1)
        historical-input #(-> history (nth %) :input)
        next (case direction :backward dec :forward inc)
        at-first? (= idx history-end)
        at-last?  (= idx 0)
        going-forward? (= direction :forward)
        going-backward? (= direction :backward)
        not-cycling? (-> cs :selected-history nil?)]
    (cond
      (empty? history)                   {:current current :selected-history nil}
      (and at-first? going-forward?)     {:current orig :selected-history nil}
      (and at-last? going-backward?)     nil ; do nothing
      (and not-cycling? going-forward?)  nil ; do nothing
      (and not-cycling? going-backward?) {:current (historical-input history-end)
                                          :selected-history {:orig current
                                                             :index history-end}}
      :else                              {:current (historical-input (next idx))
                                          :selected-history {:orig orig
                                                             :index (next idx)}})))

(defn cycle-suggestions [cmd direction]
  (let [{:keys [options selected]} (cmd :suggestions)
        next (case direction :backward dec :forward inc)
        nopts (count options)
        next-selected (cond
                        (nil? selected)     (case direction
                                                  :backward (- nopts 1)
                                                  :forward 0)
                        (>= (next selected) nopts) 0
                        (< (next selected) 0) (- nopts 1)
                        :else               (next selected))]
    (when (not-empty options)
      {:suggestions {:options options :selected next-selected}})))

(def null-suggestions {:options nil :selected nil})

(defn complete-suggestion [{:keys [suggestions]}]
  (let [{:keys [options selected]} suggestions]
    {:current (nth options selected) :suggestions null-suggestions}))

(re-frame/reg-event-db
 ::hover-suggestion
 (re-frame/path :cmdline)
 (fn [cmd [_ idx]]
     (assoc-in cmd [:suggestions :selected] idx)))

(re-frame/reg-event-fx
 ::click-suggestion
 (fn [cofx [_ prompt cmd]]
   (let [db (cofx :db)]
     {:db (assoc-in db [:cmdline :current] cmd)
      :fx [[:dispatch [::submit-cmd prompt]]]})))

(re-frame/reg-event-db
 ::prompt-keypress
 (re-frame/path :cmdline)
 (fn [cmd [_ event]]
   (let [direction (case (.-key event)
                     "ArrowUp"   :backward
                     "ArrowDown" :forward
                     "ArrowRight" :right
                     nil)
         tab (= (.-key event) "Tab")
         escape (= (.-key event) "Escape")]
     (cond
      (contains? #{:backward :forward} direction) (do (.preventDefault event)
                                                      (if (some-> cmd :suggestions :options)
                                                          (merge cmd (cycle-suggestions cmd direction))
                                                          (merge cmd (cycle-history cmd direction))))
      (and (= :right direction)
           (some-> cmd :suggestions :selected)) (do (.preventDefault event)
                                                    (merge cmd (complete-suggestion cmd)))
      (and tab
           (= 1 (some-> cmd :suggestions :options count))
           (some-> cmd :suggestions :selected)) (do (.preventDefault event)
                                                    (merge cmd (complete-suggestion cmd)))
      tab (do (.preventDefault event)
              (merge cmd (cycle-suggestions cmd :forward)))
      (and escape (some-> cmd :suggestions :options)) (do (.preventDefault event)
                                                          (merge cmd {:suggestions null-suggestions}))
      :else     cmd))))

(re-frame/reg-event-fx
 ::submit-cmd
 (fn [cofx [_ prompt]]
  (let [db (cofx :db)
        cmdline (some-> db :cmdline :current)]
    {:db
     (-> db
      (assoc-in [:cmdline :current] nil)
      (assoc-in [:cmdline :selected-history] nil)
      (assoc-in [:cmdline :suggestions] {:options nil :selected nil})
      (execute cmdline))

     ; dispatch-later to ensure that the DOM has already been updated,
     ; so that scrollMaxY has been changed appropriately.
     :fx [[:dispatch-later {:ms 10 :dispatch [::scroll-to prompt]}]]})))
